from typing import Optional
import torch
from torch import nn
from pi_vae_pytorch.layers import GINBlock, NFlowLayer, PermutationLayer


class GINFlowDecoder(nn.Module):
    """
    Define mean(p(x|z)) using GIN volume preserving flow.

    Parameters
    ----------
    - x_dim (int) - observed x dimension (should be much larger than z_dim)
    - z_dim (int) - latent z dimension
    - n_gin_blocks (int, default=2) - number of GIN blocks
    - gin_block_depth (int, default=2) - number of AffineCouplingLayers per GINBlock
    - affine_input_layer_slice_dim (int, default=None) - index at which to split an n-dimensional sample x input to each AffineCouplingLayer
    - affine_n_hidden_layers (int, default=2) - number of each AffineCouplingLayer's MLP hidden layers
    - affine_hidden_layer_dim (int, default=None) - dimension of each AffineCouplingLayer's MLP hidden layers
    - affine_hidden_layer_activation (nn.Module, default=nn.ReLU) - activation function applied to each AffineCouplingLayer's MLP hidden layers
    - nflow_n_hidden_layers (int, default=2) - number of the NFlowLayer's MLP hidden layers
    - nflow_hidden_layer_dim (int, default=None) - dimension of the NFlowLayer's MLP hidden layers
    - nflow_hidden_layer_activation (nn.Module, default=nn.ReLU) - activation function applied to the NFlowLayer's MLP hidden layers
    - observation_model (str, default=poisson) - poisson or gaussian. Default: poisson

    Notes
    -----
    - affine_input_layer_slice_dim (int) - when None, x_dim // 2 is assigned. Otherwise assigns the specified value, assuming 0 < affine_input_layer_slice_dim < x_dim.
    - affine_hidden_layer_dim (int) - when None, x_dim // 4 is assigned. Otherwise max(affine_hidden_layer_dim, x_dim // 4).
    - nflow_hidden_layer_dim (int) - when None, x_dim // 4 is assigned. Otherwise max(nflow_hidden_layer_dim, x_dim // 4).
    """

    def __init__(
        self,
        x_dim: int,
        z_dim: int,
        n_gin_blocks: int = 2,
        gin_block_depth: int = 2,
        affine_input_layer_slice_dim: Optional[int] = None,
        affine_n_hidden_layers: int = 2,
        affine_hidden_layer_dim: Optional[int] = None,
        affine_hidden_layer_activation: nn.Module = nn.ReLU,
        nflow_n_hidden_layers: int = 2,
        nflow_hidden_layer_dim: Optional[int] = None,
        nflow_hidden_layer_activation: nn.Module = nn.ReLU,
        observation_model: str = 'poisson'
        ) -> None:
        super().__init__()

        self.output_activation = nn.Softplus() if observation_model == 'poisson' else nn.Identity()

        self.n_flow = NFlowLayer(
            x_dim=x_dim,
            z_dim=z_dim,
            n_hidden_layers=nflow_n_hidden_layers,
            hidden_layer_dim=nflow_hidden_layer_dim,
            hidden_layer_activation=nflow_hidden_layer_activation
        )

        layers = []
        for _ in range(n_gin_blocks):
            layers.append(PermutationLayer(x_dim))
            layers.append(
                GINBlock(
                    x_dim=x_dim,
                    n_affine_layers=gin_block_depth,
                    affine_input_layer_slice_dim=affine_input_layer_slice_dim,
                    affine_n_hidden_layers=affine_n_hidden_layers,
                    affine_hidden_layer_dim=affine_hidden_layer_dim,
                    affine_hidden_layer_activation=affine_hidden_layer_activation
                )
            )

        self.gin_blocks = nn.Sequential(*layers)
    
    def forward(
        self,
        z: torch.Tensor
        ) -> torch.Tensor:
        """
        Projects latent z into observation dimension (x_dim).
        """
       
        output = self.n_flow(z)
        output = self.gin_blocks(output)
        # Softplus or Identity
        return self.output_activation(output)



from typing import Optional
import torch
from torch import nn


class ELBOLoss(nn.Module):
    """ 
    pi-VAE Loss function which maximizes the evidence lower bound (ELBO) of p(x | u)

    Parameters
    ----------
    - version (int, default=2) - version of the loss function
    - alpha (float, default=0.5) - Weights the contribution of the encoder KL loss and posterior KL loss to the total KL loss
    - observation_model (str, default=poisson) - poisson or gaussian
    - device (torch.device, default=None) - object representing the device on which a Tensor will be allocated 
    
    Inputs
    ------
    - x (Tensor) - observed x. Size([n_samples, x_dim])
    - posterior_firing_rate (Tensor) - decoded firing rates. Size([n_samples, x_dim])
    - posterior_mean (Tensor) - means from posterior q(z|x,u)~q(z|x)p(z|u). Size([n_samples. z_dim])
    - posterior_log_variance (Tensor) - log of variances from posterior q(z|x,u)~q(z|x)p(z|u). Size([n_samples, z_dim])
    - label_mean (Tensor) - means from label prior p(z|u). Size([n_samples, z_dim])
    - label_log_variance (Tensor) - log of variances from label prior p(z|u). Size([n_samples, z_dim])
    - encoder_mean (Tensor, default=None) - means from encoder p(z|x). Size([n_samples, z_dim])
    - encoder_log_variance (Tensor, default=None) - means from encoder p(z|x). Size([n_samples, z_dim])
    - observation_noise_model (nn.Module, default=None) - if gaussian observation model, set the observation noise level as different real numbers.

    Returns
    -------
    - (Tensor) - The total loss. Size([])

    Notes
    -----
    observation_noise_model (nn.Module) - PiVAE uses an nn.Linear(1, x_dim, bias=False) Module

    min -log p(x|z) + E_q log(q(z))-log(p(z|u))
    cross entropy
    q (mean1, var1) p (mean2, var2)
    E_q log(q(z))-log(p(z|u)) = -0.5*(1-log(var2/var1) - (var1+(mean2-mean1)^2)/var2)
    E_q(z|x,u) log(q(z|x,u))-log(p(z|u)) = -0.5*(log(2*pi*var2) + (var1+(mean2-mean1)^2)/var2)
    p(z) = q(z|x) = N(f(x), g(x)) parametrized by nn
    """

    def __init__(
        self,
        version: int = 2,
        alpha: float = 0.5,
        observation_model: str = 'poisson',
        device: Optional[torch.device] = None
        ) -> None:
        super().__init__()

        if version == 1:
            self.version = version
        elif version == 2:
            self.version = version

            if alpha >= 0 and alpha <= 1:
                self.alpha = alpha
            else:
                raise ValueError(f"Alpha outside of allowed interval [0, 1].")
        else:
            raise ValueError(f"Invalid loss version.")
        
        if observation_model == 'gaussian':
           self.ones = torch.ones(size=(1, 1), device=device)
        elif observation_model != 'poisson' and observation_model != 'gaussian':
            raise ValueError(f"Invalid observation_model: {observation_model}")
        
        self.observation_model = observation_model

    def forward(
        self,
        x: torch.Tensor,
        posterior_firing_rate: torch.Tensor,
        posterior_mean: torch.Tensor, 
        posterior_log_variance: torch.Tensor,
        label_mean: torch.Tensor, 
        label_log_variance: torch.Tensor,
        encoder_mean: Optional[torch.Tensor] = None,
        encoder_log_variance: Optional[torch.Tensor] = None,
        observation_noise_model: Optional[nn.Module] = None
        ) -> torch.Tensor:

        if self.observation_model == 'poisson':
            observation_log_liklihood = torch.sum(posterior_firing_rate - x * torch.log(posterior_firing_rate), dim=-1)
        else:
            observation_log_variance = observation_noise_model(self.ones)
            observation_log_liklihood = torch.sum(torch.square(posterior_firing_rate - x) / (2 * torch.exp(observation_log_variance)) + (observation_log_variance / 2), dim=-1)

        if self.version == 1:
            kl_loss = 0.5 * self.compute_kl_loss(posterior_mean, posterior_log_variance, label_mean, label_log_variance)
        else:
            encoder_kl_loss = self.compute_kl_loss(encoder_mean, encoder_log_variance, label_mean, label_log_variance)
            posterior_kl_loss = self.compute_kl_loss(posterior_mean, posterior_log_variance, label_mean, label_log_variance)
            kl_loss = self.alpha * encoder_kl_loss + (1 - self.alpha) * posterior_kl_loss
        
        return torch.mean(observation_log_liklihood - kl_loss)
    
    @staticmethod
    def compute_kl_loss(
        mean_0: torch.Tensor,
        log_variance_0: torch.Tensor,
        mean_1: torch.Tensor,
        log_variance_1: torch.Tensor
        ) -> torch.Tensor:
        """
        Computes the Kullback–Leibler divergence between two distributions.

        Parameters
        ----------
        - mean_0 (Tensor) - predicted mean(s) of a distribution. Size([n_samples, sample_dim])
        - log_variance_0 (Tensor) - predicted log of variance(s) of a distribution. Size([n_samples, sample_dim])
        - mean_1 (Tensor) - predicted mean(s) of a distribution. Size([n_samples, sample_dim])
        - log_variance_1 (Tensor) - predicted log of variance(s) of a distribution. Size([n_samples, sample_dim])

        Returns
        -------
        - (Tensor) - the KL divergence loss. Size([])
        """

        kl_loss = 1 + log_variance_0 - log_variance_1 - ((torch.square(mean_0 - mean_1) + torch.exp(log_variance_0)) / torch.exp(log_variance_1))
        return 0.5 * torch.sum(kl_loss, dim=-1)



from typing import Union, Optional
import torch
from torch import nn
from pi_vae_pytorch.decoders import GINFlowDecoder
from pi_vae_pytorch.encoders import MLPEncoder
from pi_vae_pytorch.label_prior import LabelPriorContinuous, LabelPriorDiscrete


class PiVAE(nn.Module):
    """
    The pi-VAE model which constructs a latent variable model of neural activity while 
    simultaneously modeling the relation between the latent and task variables.

    Parameters
    ----------
    - x_dim (int) - observed x dimension
    - u_dim (int) -  label u dimension
    - z_dim (int) - latent z dimension
    - discrete_labels (bool, default=True) - label u format discrete(True) or continuous(False)
    - encoder_n_hidden_layers (int, default=2) - number of the MLPEncoder's MLP hidden layers
    - encoder_hidden_layer_dim (int, default=128) - dimension of the MLPEncoder's MLP hidden layers
    - encoder_hidden_layer_activation (nn.Module, default=nn.Tanh) - activation function applied to the MLPEncoder's MLP hidden layers
    - decoder_n_gin_blocks (int, default=2) - number of GINBlocks in the GINFlowDecoder
    - decoder_gin_block_depth (int, default=2) - depth of each GINBlock in the GINFlowDecoder
    - decoder_affine_input_layer_slice_dim (int, default=None) - index at which to split an n-dimensional sample x input to each AffineCouplingLayer
    - decoder_affine_n_hidden_layers (int, default=2) - number of each AffineCouplingLayer's MLP hidden layers
    - decoder_affine_hidden_layer_dim (int, default=None) - dimension of each AffineCouplingLayer's MLP hidden layers
    - decoder_affine_hidden_layer_activation (nn.Module, default=nn.ReLU) - activation function applied to each AffineCouplingLayer's MLP hidden layers
    - decoder_nflow_n_hidden_layers (int, default=2) - number of the NFlowLayer's MLP hidden layers
    - decoder_nflow_hidden_layer_dim (int, default=None) - dimension of the NFlowLayer's MLP hidden layers
    - decoder_nflow_hidden_layer_activation (nn.Module, default=nn.ReLU) - activation function applied to the NFlowLayer's MLP hidden layers
    - decoder_observation_model (str, default=poisson) - GINFlowDecoder's observation model poisson/gaussian
    - decoder_fr_clamp_min (float, default=1E-7) - min value used when clamping decoded firing rates
    - decoder_fr_clamp_max (float, default=1E7) - max value used when clamping decoded firing rates
    - label_prior_n_hidden_layers (int, default=2) - number of the LabelPriorContinuous's MLP hidden layers
    - label_prior_hidden_layer_dim (int, default=32) - dimension of the LabelPriorContinuous's MLP hidden layers
    - label_prior_hidden_layer_activation (nn.Module, default=nn.Tanh) - activation function applied to the LabelPriorContinuous's MLP hidden layers

    Notes
    -----
    - decoder_affine_input_layer_slice_dim (int) - when None, x_dim // 2 is assigned. Otherwise assigns the specified value, assuming 0 < decoder_affine_input_layer_slice_dim < x_dim.
    - decoder_affine_hidden_layer_dim (int) - when None, x_dim // 4 is assigned. Otherwise max(decoder_affine_hidden_layer_dim, x_dim // 4).
    - decoder_nflow_hidden_layer_dim (int) - when None, x_dim // 4 is assigned. Otherwise max(decoder_nflow_hidden_layer_dim, x_dim // 4).
    """
    
    def __init__(
        self,
        x_dim: int,
        u_dim: int,
        z_dim: int,
        discrete_labels: bool = True,
        encoder_n_hidden_layers: int = 2,
        encoder_hidden_layer_dim: int = 128,
        encoder_hidden_layer_activation: nn.Module = nn.Tanh,
        decoder_n_gin_blocks: int = 2,
        decoder_gin_block_depth: int = 2,
        decoder_affine_input_layer_slice_dim: Optional[int] = None,
        decoder_affine_n_hidden_layers: int = 2,
        decoder_affine_hidden_layer_dim: Optional[int] = None,
        decoder_affine_hidden_layer_activation: nn.Module = nn.ReLU,
        decoder_nflow_n_hidden_layers: int = 2,
        decoder_nflow_hidden_layer_dim: Optional[int] = None,
        decoder_nflow_hidden_layer_activation: nn.Module = nn.ReLU,
        decoder_observation_model: str = 'poisson',
        decoder_fr_clamp_min: float = 1E-7,
        decoder_fr_clamp_max: float = 1E7,
        label_prior_n_hidden_layers: int = 2,
        label_prior_hidden_layer_dim: int = 32,
        label_prior_hidden_layer_activation: nn.Module = nn.Tanh
        ) -> None:
        super().__init__()

        if decoder_observation_model == 'gaussian':
            self.observation_noise_model = nn.Linear(
                in_features=1,
                out_features=x_dim, 
                bias=False
            )
        elif decoder_observation_model == 'poisson':
            self.observation_noise_model = None
        else:
            raise ValueError(f"Invalid observation model: {decoder_observation_model}")
        
        self.decoder_observation_model = decoder_observation_model
        
        self.decoder = GINFlowDecoder(
            x_dim=x_dim,
            z_dim=z_dim,
            n_gin_blocks=decoder_n_gin_blocks,
            gin_block_depth=decoder_gin_block_depth,
            affine_input_layer_slice_dim=decoder_affine_input_layer_slice_dim,
            affine_n_hidden_layers=decoder_affine_n_hidden_layers,
            affine_hidden_layer_dim=decoder_affine_hidden_layer_dim,
            affine_hidden_layer_activation=decoder_affine_hidden_layer_activation,
            nflow_n_hidden_layers=decoder_nflow_n_hidden_layers,
            nflow_hidden_layer_dim=decoder_nflow_hidden_layer_dim,
            nflow_hidden_layer_activation=decoder_nflow_hidden_layer_activation,
            observation_model=decoder_observation_model
        )

        if decoder_fr_clamp_min < decoder_fr_clamp_max:
            self.decoder_fr_clamp_min = decoder_fr_clamp_min
            self.decoder_fr_clamp_max = decoder_fr_clamp_max
        else:
            raise ValueError(f"decoder_fr_clamp_min: {decoder_fr_clamp_min} must be less than decoder_fr_clamp_max: {decoder_fr_clamp_max}")
        
        
        self.encoder = MLPEncoder(
            x_dim=x_dim,
            z_dim=z_dim,
            n_hidden_layers=encoder_n_hidden_layers,
            hidden_layer_dim=encoder_hidden_layer_dim,
            activation=encoder_hidden_layer_activation
        )
        
        if discrete_labels:
            self.label_prior = LabelPriorDiscrete(
                u_dim=u_dim,
                z_dim=z_dim
            )
        else:
            self.label_prior = LabelPriorContinuous(
                u_dim=u_dim,
                z_dim=z_dim,
                n_hidden_layers=label_prior_n_hidden_layers,
                hidden_layer_dim=label_prior_hidden_layer_dim,
                hidden_layer_activation=label_prior_hidden_layer_activation
            )

        self.inference = False

    def forward(
        self,
        x: torch.Tensor,
        u: Optional[torch.Tensor] = None
        ) -> dict[str, torch.Tensor]:

        # Encode each sample observation x to latent z approximating q(z|x)
        encoder_mean, encoder_log_variance = self.encoder(x)

        with torch.no_grad():
            # Sample latent z using reparameterization trick
            encoder_z_sample = self.reparameterization_trick(mean=encoder_mean, log_variance=encoder_log_variance)

            # Generate firing rate using sampled latent z
            encoder_firing_rate = self.decoder(encoder_z_sample)

            if self.decoder_observation_model == 'poisson':
                encoder_firing_rate = torch.clamp(encoder_firing_rate, min=self.decoder_fr_clamp_min, max=self.decoder_fr_clamp_max)
        
        if self.inference:
            return {
                'encoder_firing_rate': encoder_firing_rate,
                'encoder_z_sample': encoder_z_sample,
                'encoder_mean': encoder_mean,
                'encoder_log_variance': encoder_log_variance                
            }
        else:
            # Mean and log of variance for each sample using label prior p(z|u)
            label_mean, label_log_variance = self.label_prior(u)

            # Compute the full posterior of q(z|x,u)~q(z|x)p(z|u) as a product of Gaussians
            posterior_mean, posterior_log_variance = self.compute_posterior(encoder_mean, encoder_log_variance, label_mean, label_log_variance)  

            # Sample latent z using reparameterization trick
            posterior_z_sample = self.reparameterization_trick(mean=posterior_mean, log_variance=posterior_log_variance)

            # Generate firing rate using sampled latent z
            posterior_firing_rate = self.decoder(posterior_z_sample)
            if self.decoder_observation_model == 'poisson':
                posterior_firing_rate = torch.clamp(posterior_firing_rate, min=self.decoder_fr_clamp_min, max=self.decoder_fr_clamp_max)

            return {
                'encoder_firing_rate': encoder_firing_rate,
                'encoder_z_sample': encoder_z_sample,
                'encoder_mean':   encoder_mean,
                'encoder_log_variance': encoder_log_variance,
                'label_mean': label_mean,
                'label_log_variance': label_log_variance,
                'posterior_firing_rate': posterior_firing_rate,
                'posterior_z_sample': posterior_z_sample,
                'posterior_mean': posterior_mean,
                'posterior_log_variance': posterior_log_variance
            }

    def decode(
        self,
        z: torch.Tensor
        ) -> torch.Tensor:
        """
        Projects samples in the model's latent space (`z_dim`) into the model's observation space (`x_dim`) by passing them through the model's decoder module.

        Parameters
        ----------
        - `z` (Tensor) - sample(s) the model's latent space (`z_dim`) to be decoded. `Size([n_samples, z_dim])`

        Returns
        -------
        - `decoded` (Tensor) - sample(s) in the model's observation space (`x_dim`). `Size([n_samples, x_dim])` 
        """

        with torch.no_grad():
            decoded = self.decoder(z)

            if self.decoder_observation_model == 'poisson':
                decoded = torch.clamp(decoded, min=self.decoder_fr_clamp_min, max=self.decoder_fr_clamp_max)

        return decoded

    def encode(
        self,
        x: torch.Tensor,
        return_stats: bool = False
        ) -> Union[torch.Tensor, tuple[torch.Tensor, torch.Tensor, torch.Tensor]]:
        """
        Projects samples in the model's observation space (`x_dim`) into the model's latent space (`z_dim`) by passing them through the model's encoder module.

        Parameters
        ----------
        - `x` (Tensor) - the sample(s) in the model's observation space (`x_dim`) to be encoded. `Size([n_samples, x_dim])`
        - `return_stats` (bool, default=False) - if `True`, the mean and log of variance associated with the encoded sample are returned; otherwise only the encoded sample is returned.

        Returns
        -------
        - `encoded` (Tensor) - sample(s) in the model's latent space(`z_dim`). `Size([n_samples, z_dim])`
        - `encoded_mean` (Tensor) [optional] - mean(s) of the encoded sample(s). `Size([n_samples, z_dim])`
        - `encoded_log_variance` (Tensor) [optional] - log of variance(s) of the encoded sample(s). `Size([n_samples, z_dim])`
        """

        with torch.no_grad():
            encoded_mean, encoded_log_variance = self.encoder(x)
            # Sample latent z using reparameterization trick
            encoded = self.reparameterization_trick(mean=encoded_mean, log_variance=encoded_log_variance)

        if return_stats:
            return encoded, encoded_mean, encoded_log_variance
        else:
            return encoded

    def get_label_statistics(
        self,
        u: Union[float, int, list, tuple, torch.Tensor],
        device: Optional[torch.device] = None
        ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Returns the mean and log of the variance associated with label `u`.

        Parameters
        ----------
        - `u` (int, float, list, tuple, or Tensor) - the label of the generated samples
        - `device` (torch.device, default=None) - the torch device on which the model resides

        Returns
        -------
        - `label_mean` (Tensor) - mean associated with the specified label. `Size([1, z_dim])`
        - `label_log_variance` (Tensor) - log of the variance associated with  the specified label. `Size([1, z_dim])`
        """

        with torch.no_grad():
            if isinstance(u, int): # discrete label
                u = torch.as_tensor([u], device=device)
            elif isinstance(u, float): # continuous label
                u = torch.as_tensor([u], device=device).unsqueeze(dim=0)
            elif isinstance(u, list) or isinstance(u, tuple): # continuous label
                u = torch.as_tensor(u, dtype=torch.float, device=device).unsqueeze(dim=0)

            # Mean and log of variance of label u using label prior estimator of p(z|u)
            label_mean, label_log_variance = self.label_prior(u) 
        
        return label_mean, label_log_variance
    
    def predict_labels(
        self,
        x: torch.Tensor,
        n_samples: int = 1000,
        device: Optional[torch.device] = None
    ) -> torch.Tensor:
        """
        Returns the label probabilities associated with observation(s) `x`.

        Parameters
        ----------
        - `x` (Tensor) - data in the model's observation space (`x_dim`). `Size([n_observations, x_dim])`
        - `n_samples` (int, default=1000) - number of samples drawn per label during Monte Carlo estimation
        - `device` (torch.device, default=None) - the torch device on which the model resides

        Returns
        -------
        - `predictions` (Tensor) - the label probabilities of the observation(s). `Size([n_observations, u_dim])`
        """

        with torch.no_grad():
            n_labels = self.label_prior.embedded_mean.weight.shape[0]
            label_means, label_stds = None, None

            z = self.encode(x)
            _, latent_dim = z.shape

            for u in range(n_labels):
                mean, log_variance = self.get_label_statistics(u=u, device=device)

                if label_means is None:
                    label_means, label_stds = mean, torch.exp(0.5 * log_variance)
                else:
                    label_means = torch.cat((label_means, mean), dim=0)
                    label_stds = torch.cat((label_stds, torch.exp(0.5 * log_variance)), dim=0)

            # Sample from p(z|u) for each class
            z_samples = label_means.unsqueeze(dim=0) + label_stds.unsqueeze(dim=0) * torch.randn(n_samples, n_labels, latent_dim)

            # Compute distances between sampled z and actual z
            z = z.unsqueeze(1).unsqueeze(dim=0)  # Size([1, batch_size, 1, latent_dim])
            distances = torch.square(z - z_samples.unsqueeze(1)).sum(dim=-1)  # Size([n_samples, batch_size, n_labels])

            # Estimate p(z|u) using Monte Carlo (negative distances as proxy for likelihood)
            proxy_log_liklihood = torch.neg(distances.mean(dim=0))  # Average over samples, Size([batch_size, n_labels])
            
            # Normalize with softmax to approximate p(u|z)
            return torch.nn.functional.softmax(proxy_log_liklihood, dim=1)

    def sample(
        self,
        u: Union[float, int, list, tuple, torch.Tensor],
        n_samples: int = 1,
        return_z: bool = False,
        device: Optional[torch.device] = None
        ) -> torch.Tensor:
        """
        Generates samples in the model's x dimension using a specified label u.

        Parameters
        ----------
        - u (int, float, list, tuple, or Tensor) - the label of the generated samples
        - n_samples (int, default=1) - the number of samples to generate
        - return_z (bool, default=False) - if True the latent space samples are also returned
        - device (torch.device, default=None) - the torch device on which the model resides

        Returns
        -------
        - samples (Tensor) - the projected latent space samples corresponding to the specified label. Size([n_samples, x_dim])
        - z_samples (Tensor) [optional] - the generated latent space samples corresponding to the specified label. Size([n_samples, z_dim]) 
        """
        with torch.no_grad():
            if isinstance(u, int): # discrete label
                u = torch.as_tensor([u], device=device)
            elif isinstance(u, float): # continuous label
                u = torch.as_tensor([u], device=device).unsqueeze(dim=0)
            elif isinstance(u, list) or isinstance(u, tuple): # continuous label
                u = torch.as_tensor(u, dtype=torch.float, device=device).unsqueeze(dim=0)
            
            # Mean and log of variance of label u
            mean, log_variance = self.label_prior(u)

            # Create covariance matrix
            variance = torch.exp(log_variance).squeeze(dim=0)
            covariance = torch.diag_embed(variance)

            # Create distribution used for sampling
            distribution = torch.distributions.multivariate_normal.MultivariateNormal(mean, covariance)
            
            # Generate z samples
            z_samples = distribution.sample((n_samples,))
            z_samples = z_samples.squeeze(dim=1)

            # Lift samples to x dimension
            samples = self.decoder(z_samples)

        if return_z:
            return samples, z_samples
        else:
            return samples

    def sample_z(
        self,
        u: Union[float, int, list, tuple, torch.Tensor],
        n_samples: int = 1,
        device: Optional[torch.device] = None
        ) -> torch.Tensor:
        """
        Generates samples in the model's z dimension using a specified label u.

        Parameters
        ----------
        - u (int, float, list, tuple, or Tensor) - the label of the generated samples
        - n_samples (int, default=1) - the number of samples to generate
        - device (torch.device, default=None) - the torch device on which the model resides

        Returns
        -------
        - samples (Tensor) - the generated samples corresponding to the specified label. Size([n_samples, z_dim]) 
        """

        with torch.no_grad():
            if isinstance(u, int): # discrete label
                u = torch.as_tensor([u], device=device)
            elif isinstance(u, float): # continuous label
                u = torch.as_tensor([u], device=device).unsqueeze(dim=0)
            elif isinstance(u, list) or isinstance(u, tuple): # continuous label
                u = torch.as_tensor(u, dtype=torch.float, device=device).unsqueeze(dim=0)

            # Mean and log of variance of label u
            mean, log_variance = self.label_prior(u)

            # Create covariance matrix
            variance = torch.exp(log_variance).squeeze(dim=0)
            covariance = torch.diag_embed(variance)

            # Create distribution used for sampling
            distribution = torch.distributions.multivariate_normal.MultivariateNormal(mean, covariance)
                
            # Generate z samples
            samples = distribution.sample((n_samples,))
            samples = samples.squeeze(dim=1)                          

        return samples
    
    def set_inference_mode(
        self,
        state: bool
        ) -> None:
        """
        Toggles the model's inference state flag. When `True`, the model's `forward` method does not utilize the `u` parameter. When `False`,  the `u` parameter is utilized.

        Parameters
        ----------
        - `state` (bool) - the desired inference state

        Returns
        -------
        - `None`
        """

        self.inference = state

    @staticmethod
    def compute_posterior(
    mean_0: torch.Tensor, 
    log_variance_0: torch.Tensor, 
    mean_1: torch.Tensor, 
    log_variance_1: torch.Tensor
    ) -> torch.Tensor:
        """
        Computes the posterior of two distributions as a product of Gaussians.

        Parameters
        ----------
        - mean_0 (Tensor) - mean of a distribution. Size([n_samples, sample_dim])
        - log_variance_0 (Tensor) - log of variance of a distribution. Size([n_samples, sample_dim])
        - mean_1 (Tensor) - mean of a distribution. Size([n_samples, sample_dim])
        - log_variance_1 (Tensor) - log of variance of a distribution. Size([n_samples, sample_dim])

        Returns
        -------
        - posterior_mean: mean of the posterior. Size([n_samples, sample_dim])
        - posterior_log_variance: log of variances of the posterior. Size([n_samples, sample_dim])

        Notes
        -----
        - The sample_dim should be equivalent amongst all four parameters. 
        """

        variance_difference = log_variance_0 - log_variance_1
        posterior_mean = (mean_0 / (1 + torch.exp(variance_difference))) + (mean_1 / (1 + torch.exp(torch.neg(variance_difference))))
        posterior_log_variance = log_variance_0 + log_variance_1 - torch.log(torch.exp(log_variance_0) + torch.exp(log_variance_1))

        return posterior_mean, posterior_log_variance
    
    @staticmethod
    def reparameterization_trick(
    mean: torch.Tensor, 
    log_variance: torch.Tensor
    ) -> torch.Tensor:
        """
        Samples from an isotropic unit Gaussian distribution using the reparameterization trick.

        Parameters
        ----------
        - mean (Tensor) - mean of a distribution. Size([n_samples, sample_dim])
        - log_variance (Tensor) - log of variance of a distribution. Size([n_samples, sample_dim])

        Returns
        -------
        - samples: a sample from the distribution. Size([n_samples, sample_dim])
        """

        return mean + torch.exp(0.5 * log_variance) * torch.randn_like(mean)



import torch
from torch import nn
from pi_vae_pytorch.layers import MLP


class LabelPriorContinuous(nn.Module):
    """
    Compute the mean and log of variance of label prior p(z|u) for continuous label u.

    Parameters
    ----------
    - u_dim (int) - label u dimension
    - z_dim (int) - latent z dimension
    - n_hidden_layers (int) - number of MLP hidden layers. Default: 2
    - hidden_layer_dim (int) - dimension of MLP hidden layers. Default: 32
    - hidden_layer_activation (nn.Module) - activation function applied to MLP hidden layers. Default: nn.Tanh
    """

    def __init__(
        self,
        u_dim: int,
        z_dim: int,
        n_hidden_layers: int = 2,
        hidden_layer_dim: int = 32,
        hidden_layer_activation: nn.Module = nn.Tanh
        ) -> None:
        super().__init__()

        self.net = MLP(
            in_features=u_dim,
            out_features=z_dim*2,
            n_hidden_layers=n_hidden_layers,
            hidden_layer_features=hidden_layer_dim,
            activation=hidden_layer_activation
        )
    
    def forward(
        self,
        u: torch.Tensor
        ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Maps u to mean and log of variance of p(z|u).
        """
        
        z_prior = self.net(u)
        # lambda_mean, lambda_log_variance
        return torch.chunk(input=z_prior, chunks=2, dim=-1)
    

class LabelPriorDiscrete(nn.Module):
    """
    Compute the mean and log of variance of label prior p(z|u) for discrete label u.

    Parameters
    ----------
    - u_dim (int) - label u dimension
    - z_dim (int) - latent z dimension
    """

    def __init__(
        self,
        u_dim: int,
        z_dim: int
        ) -> None:
        super().__init__()

        self.embedded_mean = nn.Embedding(
            num_embeddings=u_dim,
            embedding_dim=z_dim
        )

        self.embedded_log_variance = nn.Embedding(
            num_embeddings=u_dim,
            embedding_dim=z_dim
        )

    def forward(
        self,
        u: torch.Tensor
        ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Maps u to mean and log of variance of p(z|u).
        """

        # lambda_mean, lambda_log_variance
        return self.embedded_mean(u), self.embedded_log_variance(u)



import torch
from torch import nn
from pi_vae_pytorch.layers import MLP


class MLPEncoder(nn.Module):
    """
    Defines mean and log of variance of q(z|x).

    Parameters
    ----------
    - x_dim (int) - observed x dimension
    - z_dim (int) - latent z dimension
    - n_hidden_layers (int, default=2) - number of MLP hidden layers
    - hidden_layer_dim (int, default=128) - dimension of each MLP hidden layer
    - activation (nn.Module, default=nn.Tanh) - activation function applied to each MLP hidden layer
    """

    def __init__(
        self,
        x_dim: int,
        z_dim: int,
        n_hidden_layers: int = 2,
        hidden_layer_dim: int = 128,
        activation: nn.Module = nn.Tanh
        ) -> None:
        super().__init__()

        self.net = MLP(
            in_features=x_dim,
            out_features=z_dim*2,
            n_hidden_layers=n_hidden_layers,
            hidden_layer_features=hidden_layer_dim,
            activation=activation
        )
    
    def forward(
        self,
        x: torch.Tensor
        ) -> tuple[torch.Tensor, torch.Tensor]:
        """
        Maps observed x to mean and log of variance of q(z|x).
        """

        q_z = self.net(x)
        # phi_mean, phi_log_variance
        return torch.chunk(input=q_z, chunks=2, dim=-1)



from typing import Optional
import torch
from torch import nn


class MLP(nn.Module):
    """
    A basic Multilayer Perceptron (MLP) module.

    Parameters
    ----------
    - in_features (int) - number of input features 
    - out_features (int) - number of output features
    - n_hidden_layers (int) - number of hidden layers
    - hidden_layer_features (int) - number of features per hidden layer
    - activation (nn.Module) - activation function applied to hidden layers
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        n_hidden_layers: int,
        hidden_layer_features: int,
        activation: nn.Module
        ) -> None:
        super().__init__()

        # Construct layer dims
        dims = [in_features]
        for _ in range(n_hidden_layers * 2):
            dims.append(hidden_layer_features)
        dims.append(out_features)

        # Build layers
        layers = []
        act = activation

        for i in range(n_hidden_layers+1):
            # Output layer
            if i == n_hidden_layers:
                act = nn.Identity
            
            layers.append(nn.Linear(dims[2*i], dims[2*i+1]))
            layers.append(act())

        self.net = nn.Sequential(*layers)

    def forward(
        self,
        input: torch.Tensor
        ) -> torch.Tensor:

        return self.net(input)


class PermutationLayer(nn.Module):
    """
    Randomly permutes n channels. 

    Parameters
    ----------
    - n_channels (int) - number of channels to permute
    """

    def __init__(
        self,
        n_channels: int
        ) -> None:
        super().__init__()

        self.register_buffer('perm', torch.randperm(n_channels))
        self.register_buffer('invperm', torch.argsort(self.perm))

    def forward(
        self,
        x: torch.Tensor
        ) -> torch.Tensor:

        return x[:, self.perm]

    def backward(
        self,
        x: torch.Tensor
        ) -> torch.Tensor:

        return x[:, self.invperm]


class NFlowLayer(nn.Module):
    """
    Define the first layer in GIN flow, which maps z to the cancatenation of z and t(z). t is parameterized by self.net. 
    This is equivalent to GIN model with input as z1:dim_z padding dim_x - dim_z zeros.

    Parameters
    ----------
    -x_dim (int) - observed x dimension
    - z_dim (int) - latent z dimension
    - n_hidden_layers (int, default=2) - number of MLP hidden layers
    - hidden_layer_dim (int, default=None) - dimension of MLP hidden layers
    - hidden_layer_activation (nn.Module, default=nn.ReLU) - activation function applied to MLP hidden layers

    Notes
    -----
    - hidden_layer_dim (int) - when None, x_dim // 4 is assigned. Otherwise max(hidden_layer_dim, x_dim // 4).
    """

    def __init__(
        self,
        x_dim: int,
        z_dim: int,
        n_hidden_layers: int = 2,
        hidden_layer_dim: Optional[int] = None,
        hidden_layer_activation: nn.Module = nn.ReLU
        ) -> None:
        super().__init__()

        # Compute hidden layer dimension
        if hidden_layer_dim is None:
            hidden_dim = x_dim // 4
        else:
            hidden_dim = max(hidden_layer_dim, x_dim // 4)

        self.net = MLP(
            in_features=z_dim,
            out_features=x_dim - z_dim,
            n_hidden_layers=n_hidden_layers,
            hidden_layer_features=hidden_dim,
            activation=hidden_layer_activation
        )

    def forward(
        self,
        z: torch.Tensor
        ) -> torch.Tensor:
        """
        Maps z to the cancatenation of z and t(z). 
        """

        t_z = self.net(z)
        return torch.cat((z, t_z), dim=-1)


class AffineCouplingLayer(nn.Module):
    """
    Performs the affine coupling transform while preserving volume by mapping input x to
    [x_{1:x_slice_dim}, x_{x_slice_dim+1:n} * exp(s(x_{1:x_slice_dim})) + t(x_{1:x_slice_dim})].

    Parameters
    ----------
    - x_dim (int) - observed x dimension
    - x_slice_dim (int, default=None) - index at which to split an n-dimensional sample x
    - n_hidden_layers (int, default=2) - number of MLP hidden layers
    - hidden_layer_dim (int, default=None) - dimension of MLP hidden layers
    - hidden_layer_activation (nn.Module, default=nn.ReLU) - activation function applied to MLP hidden layers

    Notes
    -----
    - x_slice_dim (int) - when None, x_dim // 2 is assigned. Otherwise assigns the specified value, assuming 0 < x_slice_dim < x_dim. 
    - hidden_layer_dim (int) - when None, x_dim // 4 is assigned. Otherwise max(hidden_layer_dim, x_dim // 4).
    """

    def __init__(
        self,
        x_dim: int,
        x_slice_dim: Optional[int] = None,
        n_hidden_layers: int = 2,
        hidden_layer_dim: Optional[int] = None,
        hidden_layer_activation: nn.Module = nn.ReLU
        ) -> None:
        super().__init__()

        self.x_dim = x_dim

        # Compute slice dimension
        if x_slice_dim is None:
            self.slice_dim = x_dim // 2
        else:
            if x_slice_dim >= x_dim:
                raise ValueError('x_slice_dim must be less than x_dim.')
            self.slice_dim = x_slice_dim

        # Compute hidden layer dimension
        if hidden_layer_dim is not None:
            hidden_dim = max(hidden_layer_dim, x_dim // 4)
        else:
            hidden_dim = x_dim // 4

        self.net = MLP(
            in_features=self.slice_dim,
            out_features=2 * (x_dim - self.slice_dim) - 1,
            n_hidden_layers=n_hidden_layers,
            hidden_layer_features=hidden_dim,
            activation=hidden_layer_activation
        )

    def forward(
        self,
        x: torch.Tensor
        ) -> torch.Tensor:
        """
        Perform affine coupling transform while preserving volume.
        """

        # Split input
        x_1 = torch.narrow(x, 1, 0, self.slice_dim)
        x_2 = torch.narrow(x, 1, self.slice_dim, self.x_dim - self.slice_dim)

        # Compute scale and translation
        s_t = self.net(x_1)
        s_out = torch.narrow(s_t, 1, 0, self.x_dim - self.slice_dim - 1)
        t_out = torch.narrow(s_t, 1, self.x_dim - self.slice_dim - 1, self.x_dim - self.slice_dim)

        # clamp to ensure output of s is small
        s_out = 0.1 * torch.tanh(s_out)
        # preserve volume by ensuring the last layer has sum 0
        s_preserved = torch.cat((s_out, torch.sum(torch.neg(s_out), dim=-1, keepdim=True)), dim=-1)

        # perform transformation
        transform_x = x_2 * torch.exp(s_preserved) + t_out

        return torch.cat((transform_x, x_1), dim=-1)


class GINBlock(nn.Module):
    """
    General Incompressible-flow Network (GIN). Performs a series of affine coupling transformations 
    with inputs randomly permuted before applying each affine coupling function.

    Parameters
    ----------
    - x_dim (int) - observed x dimension
    - n_affine_layers (int, default=2) - number of AffineCouplingLayers in the block
    - affine_input_layer_slice_dim (int, default=None) - index at which to split an n-dimensional sample x input to each AffineCouplingLayer
    - affine_n_hidden_layers (int, default=2) - number of MLP hidden layers within each AffineCouplingLayer
    - affine_hidden_layer_dim (int, default=None) - dimension of each MLP hidden layer within each AffineCouplingLayer
    - affine_hidden_layer_activation (nn.Module, default=nn.ReLU) - activation function applied to MLP hidden layers within each AffineCouplingLayer

    Notes
    -----
    - affine_input_layer_slice_dim (int) - when None, x_dim // 2 is assigned. Otherwise assigns the specified value, assuming 0 < affine_input_layer_slice_dim < x_dim. 
    - affine_hidden_layer_dim (int) - when None, x_dim // 4 is assigned. Otherwise max(affine_hidden_layer_dim, x_dim // 4).
    """

    def __init__(
        self,
        x_dim: int,
        n_affine_layers: int = 2,
        affine_input_layer_slice_dim: Optional[int] = None,
        affine_n_hidden_layers: int = 2,
        affine_hidden_layer_dim: Optional[int] = None,
        affine_hidden_layer_activation: nn.Module = nn.ReLU
        ) -> None:
        super().__init__()

        layers = [
            AffineCouplingLayer(
                x_dim=x_dim,
                x_slice_dim=affine_input_layer_slice_dim,
                n_hidden_layers=affine_n_hidden_layers,
                hidden_layer_dim=affine_hidden_layer_dim,
                hidden_layer_activation=affine_hidden_layer_activation
            )
        ] * n_affine_layers

        self.net = nn.Sequential(*layers)

    def forward(
        self,
        x: torch.Tensor
        ) -> torch.Tensor:

        return self.net(x)
